from __future__ import annotations

import importlib
import inspect
import os
import pathlib
import sys
import tempfile
from typing import List

import pkg_resources
from clvm_tools_rs import compile_clvm as compile_clvm_rust

from chia.types.blockchain_format.program import Program
from chia.types.blockchain_format.serialized_program import SerializedProgram
from chia.util.lock import Lockfile

compile_clvm_py = None

recompile_requested = (
    (os.environ.get("CHIA_DEV_COMPILE_CLVM_ON_IMPORT", "") != "") or ("pytest" in sys.modules)
) and os.environ.get("CHIA_DEV_COMPILE_CLVM_DISABLED", None) is None


def translate_path(p_):
    p = str(p_)
    if os.path.isdir(p):
        return p
    else:
        module_object = importlib.import_module(p)
        return os.path.dirname(inspect.getfile(module_object))


# Handle optional use of python clvm_tools if available and requested
if "CLVM_TOOLS" in os.environ:
    try:
        from clvm_tools.clvmc import compile_clvm as compile_clvm_py_candidate

        compile_clvm_py = compile_clvm_py_candidate
    finally:
        pass


def compile_clvm_in_lock(full_path: pathlib.Path, output: pathlib.Path, search_paths: List[pathlib.Path]):
    # Compile using rust (default)

    # Ensure path translation is done in the idiomatic way currently
    # expected.  It can use either a filesystem path or name a python
    # module.
    treated_include_paths = list(map(translate_path, search_paths))
    res = compile_clvm_rust(str(full_path), str(output), treated_include_paths)

    if "CLVM_TOOLS" in os.environ and os.environ["CLVM_TOOLS"] == "check" and compile_clvm_py is not None:
        # Simple helper to read the compiled output
        def sha256file(f):
            import hashlib

            m = hashlib.sha256()
            m.update(open(f).read().strip().encode("utf8"))
            return m.hexdigest()

        orig = "%s.orig" % output

        compile_clvm_py(full_path, orig, search_paths=search_paths)
        orig256 = sha256file(orig)
        rs256 = sha256file(output)

        if orig256 != rs256:
            print("Compiled original %s: %s vs rust %s\n" % (full_path, orig256, rs256))
            print("Aborting compilation due to mismatch with rust")
            assert orig256 == rs256
        else:
            print("Compilation match %s: %s\n" % (full_path, orig256))

    return res


def compile_clvm(full_path: pathlib.Path, output: pathlib.Path, search_paths: List[pathlib.Path] = []):
    with Lockfile.create(pathlib.Path(tempfile.gettempdir()) / "clvm_compile" / full_path.name):
        compile_clvm_in_lock(full_path, output, search_paths)


def load_serialized_clvm(
    clvm_filename, package_or_requirement=__name__, include_standard_libraries: bool = False, recompile: bool = True
) -> SerializedProgram:
    """
    This function takes a .clsp file in the given package and compiles it to a
    .clsp.hex file if the .hex file is missing or older than the .clsp file, then
    returns the contents of the .hex file as a `Program`.

    clvm_filename: file name
    package_or_requirement: usually `__name__` if the clvm file is in the same package
    """
    hex_filename = f"{clvm_filename}.hex"

    # Set the CHIA_DEV_COMPILE_CLVM_ON_IMPORT environment variable to anything except
    # "" or "0" to trigger automatic recompilation of the Chialisp on load.
    if recompile:
        try:
            if pkg_resources.resource_exists(package_or_requirement, clvm_filename):
                # Establish whether the size is zero on entry
                full_path = pathlib.Path(pkg_resources.resource_filename(package_or_requirement, clvm_filename))
                output = full_path.parent / hex_filename
                if not output.exists() or os.stat(full_path).st_mtime > os.stat(output).st_mtime:
                    search_paths = [full_path.parent]
                    if include_standard_libraries:
                        # we can't get the dir, but we can get a file then get its parent.
                        chia_puzzles_path = pathlib.Path(
                            pkg_resources.resource_filename(__name__, "__init__.py")
                        ).parent
                        search_paths.append(chia_puzzles_path)
                    compile_clvm(full_path, output, search_paths=search_paths)

        except NotImplementedError:
            # pyinstaller doesn't support `pkg_resources.resource_exists`
            # so we just fall through to loading the hex clvm
            pass

    clvm_hex = pkg_resources.resource_string(package_or_requirement, hex_filename).decode("utf8")
    assert len(clvm_hex.strip()) != 0
    clvm_blob = bytes.fromhex(clvm_hex)
    return SerializedProgram.from_bytes(clvm_blob)


def load_clvm(
    clvm_filename,
    package_or_requirement=__name__,
    include_standard_libraries: bool = False,
    recompile: bool = True,
) -> Program:
    return Program.from_bytes(
        bytes(
            load_serialized_clvm(
                clvm_filename,
                package_or_requirement=package_or_requirement,
                include_standard_libraries=include_standard_libraries,
                recompile=recompile,
            )
        )
    )


def load_clvm_maybe_recompile(
    clvm_filename,
    package_or_requirement=__name__,
    include_standard_libraries: bool = False,
    recompile: bool = recompile_requested,
) -> Program:
    return load_clvm(
        clvm_filename=clvm_filename,
        package_or_requirement=package_or_requirement,
        include_standard_libraries=include_standard_libraries,
        recompile=recompile,
    )


def load_serialized_clvm_maybe_recompile(
    clvm_filename,
    package_or_requirement=__name__,
    include_standard_libraries: bool = False,
    recompile: bool = recompile_requested,
) -> SerializedProgram:
    return load_serialized_clvm(
        clvm_filename=clvm_filename,
        package_or_requirement=package_or_requirement,
        include_standard_libraries=include_standard_libraries,
        recompile=recompile,
    )



import os
import pytest
import random
from itertools import permutations
from typing import List
#from hsms.streamables.program import Program
#from steprun import diag_run_clvm, compile_module_with_symbols
#from lib.program import Program
from pathlib import Path
from clvm_rs import Program
from lib.steprun import diag_run_clvm, compile_module_with_symbols
from clvm_tools_rs import get_version

print(f"clvm_tools_rs version is {get_version()}")
#include_dirs = os.getcwd()
include_dirs = [Path(__file__).parent, Path(__file__).parent.parent / "lib"]
Program.set_run_unsafe_max_cost(11000000000)

print(f"XXX: {include_dirs}")
compile_module_with_symbols(include_dirs, 'smoke_test_deep_compare.clsp')
compare_program = Program.from_bytes(bytes.fromhex(open('smoke_test_deep_compare.clvm.hex').read()))

compile_module_with_symbols(include_dirs, 'smoke_test_sort.clsp')
sort_program = Program.from_bytes(bytes.fromhex(open('smoke_test_sort.clvm.hex').read()))

compile_module_with_symbols(include_dirs, 'test_sort.clsp')
test_sort_program = Program.from_bytes(bytes.fromhex(open('test_sort.clvm.hex').read()))

compile_module_with_symbols(include_dirs, 'test_permutations.clsp')
test_permutations_program = Program.from_bytes(bytes.fromhex(open('test_permutations.clvm.hex').read()))

compile_module_with_symbols(include_dirs, 'test_reverse.clsp')
test_reverse_program = Program.from_bytes(bytes.fromhex(open('test_reverse.clvm.hex').read()))

compile_module_with_symbols(include_dirs, 'test_prepend.clsp')
test_prepend_program = Program.from_bytes(bytes.fromhex(open('test_prepend.clvm.hex').read()))

compile_module_with_symbols(include_dirs, 'test_range.clsp')
test_range_program = Program.from_bytes(bytes.fromhex(open('test_range.clvm.hex').read()))

compile_module_with_symbols(include_dirs, 'smoke_test_permutations.clsp')
smoke_test_permutations_program = Program.from_bytes(bytes.fromhex(open('smoke_test_permutations.clvm.hex').read()))

compile_module_with_symbols(include_dirs, 'test_handcalc.clsp')
test_handcalc_program = Program.from_bytes(bytes.fromhex(open('test_handcalc.clvm.hex').read()))

def as_atom_list(prg: Program) -> List[bytes]:
    """
    Pretend `prg` is a list of atoms. Return the corresponding
    python list of atoms.

    At each step, we always assume a node to be an atom or a pair.
    If the assumption is wrong, we exit early. This way we never fail
    and always return SOMETHING.
    """
    items = []
    obj = prg
    while True:
        pair = obj.pair
        if pair is None:
            break
        atom = pair[0].atom
        if atom is None:
            break
        items.append(atom)
        obj = pair[1]
    return items

def test_smoke_compare():
    compare_program.run(Program.to([]))

def test_handcalc():
    diag_run_clvm(test_handcalc_program, Program.to([]), 'test_handcalc.sym', {'print': True})

def proper_list_inner(result,cl):
    if hasattr(cl, 'pair') and cl.pair is not None:
        result.append(cl.pair[0])
        return proper_list_inner(result,cl.pair[1])
    else:
        return result

def proper_list(cl):
    result = []
    return proper_list_inner(result,cl)

def int_list(cl):
    return [Program.to(x).as_int() for x in as_atom_list(Program.to(cl))]

def de_none_list(l):
    return [x if x is not None else [] for x in l]

def with_random_lists(n,f):
    for length in range(n): # 0-10 length
        for i in range(1 + (3 * length)): # A few orders each
            orig_list = [random.randint(0,100) for x in range(length)]
            f(orig_list)

def test_prepend():
    for length1 in range(5):
        list_1 = list(range(length1))
        for length2 in range(length1):
            prepend_result = test_prepend_program.run([Program.to(list_1[:length2]),Program.to(list_1[length2:])])
            assert list_1 == int_list(prepend_result)

def test_reverse():
    def test_reverse_list(l):
        rev_args = Program.to([l])
        reversed_result = Program.to(list(reversed(l)))
        reversed_by_prog = test_reverse_program.run(rev_args)
        assert reversed_result == reversed_by_prog

    with_random_lists(10,test_reverse_list)

def test_range():
    for length in range(10):
        want_list = list(range(length))
        result = test_range_program.run(Program.to([length]))
        assert want_list == result

def do_test_permutations_of_size_n(n):
    try_list = [random.randint(0,100) for x in range(n)]
    want_set = list([list(v) for v in sorted(permutations(try_list))])
    listed_result = smoke_test_permutations_program.run(Program.to([try_list]))
    pl = proper_list(listed_result)
    perms_result = sorted([int_list(x) for x in de_none_list(pl)])
    assert want_set == perms_result

def test_permutations_0():
    do_test_permutations_of_size_n(0)

def test_permutations_1():
    do_test_permutations_of_size_n(1)

def test_permutations_2():
    n = 2
    all_a_string = 0x616161616161
    all_b_string = 0x626262626262
    for try_list in [[all_a_string,all_b_string], [all_b_string,all_a_string]]:
        want_set = list([list(v) for v in sorted(permutations(try_list))])
        listed_result = diag_run_clvm(smoke_test_permutations_program, Program.to([try_list]), 'smoke_test_permutations.sym', {'print': True})
        pl = proper_list(listed_result)
        perms_result = sorted([int_list(x) for x in de_none_list(pl)])
        assert want_set == perms_result

def test_chialisp_sort_program():
    diag_run_clvm(test_sort_program, Program.to([]), 'test_sort.sym', {'print': True})

def test_permutations_n():
    for i in range(3,6):
        do_test_permutations_of_size_n(i)

def test_chialisp_permutations_program():
    diag_run_clvm(test_permutations_program, Program.to([3, 5]), 'test_permutations.sym', {'print': True})

def test_smoke_sort():
    for length in range(7): # 0-7 length
        for i in range(1 + (3 * length)): # A few orders each
            orig_list = [random.randint(0,100) for x in range(length)]
            sort_args = Program.to([orig_list])
            sorted_list = Program.to(sorted(orig_list))
            sort_res = sort_program.run(sort_args)
            assert sort_res == sorted_list

if __name__ == '__main__':
    test_smoke_sort()



import hashlib

from hsms.streamables.program import Program

from clvm.EvalError import EvalError

noncegame = Program.from_bytes(bytes.fromhex(open("noncegame.clvm.hex").read()))
noncehash = noncegame.tree_hash()

def drun(prog: Program, args: Program):
    try:
        return prog.run(args)
    except EvalError as ee:
        print(f"brun -x -y main.sym {prog} {Program.to(args)}")
        raise

def testnonce(startnonce, maxnonce):
    for i in range(startnonce, maxnonce):
        mygame = noncegame.curry(i, noncehash)
        good_parameters = [i*2, noncegame.curry(i+1, noncehash).tree_hash(), 1, (i*4, b'g')]
        bad_parameters = [i*3, noncegame.curry(i+2, noncehash).tree_hash(), 2, (i*5, b'g')]
        assert drun(mygame, good_parameters) == b'g'
        for j in range(len(good_parameters)):
            try:
                p = list(good_parameters)
                p[j] = bad_parameters[j]
                mygame.run(p)
                assert False
            except EvalError as ee:
                pass

if __name__ == '__main__':
    testnonce(3, 7)



"""
This is an implementation of `sha256_treehash`, used to calculate
puzzle hashes in clvm.

This implementation goes to great pains to be non-recursive so we don't
have to worry about blowing out the python stack.
"""

from __future__ import annotations

from typing import Callable, List, Optional, Set

from clvm import CLVMObject

from chia.types.blockchain_format.sized_bytes import bytes32
from chia.util.hash import std_hash

Op = Callable[[List["CLVMObject"], List["Op"], Set[bytes32]], None]


def sha256_treehash(sexp: CLVMObject, precalculated: Optional[Set[bytes32]] = None) -> bytes32:
    """
    Hash values in `precalculated` are presumed to have been hashed already.
    """

    if precalculated is None:
        precalculated = set()

    def handle_sexp(sexp_stack: List[CLVMObject], op_stack: List[Op], precalculated: Set[bytes32]) -> None:
        sexp = sexp_stack.pop()
        if sexp.pair:
            p0, p1 = sexp.pair
            sexp_stack.append(p0)
            sexp_stack.append(p1)
            op_stack.append(handle_pair)
            op_stack.append(handle_sexp)
            op_stack.append(roll)
            op_stack.append(handle_sexp)
        else:
            if sexp.atom in precalculated:
                r = sexp.atom
            else:
                r = std_hash(b"\1" + sexp.atom)
            sexp_stack.append(r)

    def handle_pair(sexp_stack: List[CLVMObject], op_stack: List[Op], precalculated: Set[bytes32]) -> None:
        p0 = sexp_stack.pop()
        p1 = sexp_stack.pop()
        sexp_stack.append(std_hash(b"\2" + p0 + p1))

    def roll(sexp_stack: List[CLVMObject], op_stack: List[Op], precalculated: Set[bytes32]) -> None:
        p0 = sexp_stack.pop()
        p1 = sexp_stack.pop()
        sexp_stack.append(p0)
        sexp_stack.append(p1)

    sexp_stack = [sexp]
    op_stack: List[Op] = [handle_sexp]
    while len(op_stack) > 0:
        op = op_stack.pop()
        op(sexp_stack, op_stack, precalculated)
    return bytes32(sexp_stack[0])



import pytest
import random
from itertools import permutations
from clvm_rs import Program
from steprun import diag_run_clvm, compile_module_with_symbols

Program.set_run_unsafe_max_cost(11000000000)

compile_module_with_symbols(['.'], 'smoke_test_deep_compare.clsp')
compare_program = Program.from_bytes(bytes.fromhex(open('smoke_test_deep_compare.clvm.hex').read()))

compile_module_with_symbols(['.'], 'smoke_test_sort.clsp')
sort_program = Program.from_bytes(bytes.fromhex(open('smoke_test_sort.clvm.hex').read()))

compile_module_with_symbols(['.'], 'test_sort.clsp')
test_sort_program = Program.from_bytes(bytes.fromhex(open('test_sort.clvm.hex').read()))

compile_module_with_symbols(['.'], 'test_permutations.clsp')
test_permutations_program = Program.from_bytes(bytes.fromhex(open('test_permutations.clvm.hex').read()))

compile_module_with_symbols(['.'], 'test_reverse.clsp')
test_reverse_program = Program.from_bytes(bytes.fromhex(open('test_reverse.clvm.hex').read()))

compile_module_with_symbols(['.'], 'test_prepend.clsp')
test_prepend_program = Program.from_bytes(bytes.fromhex(open('test_prepend.clvm.hex').read()))

compile_module_with_symbols(['.'], 'test_range.clsp')
test_range_program = Program.from_bytes(bytes.fromhex(open('test_range.clvm.hex').read()))

compile_module_with_symbols(['.'], 'smoke_test_permutations.clsp')
smoke_test_permutations_program = Program.from_bytes(bytes.fromhex(open('smoke_test_permutations.clvm.hex').read()))

compile_module_with_symbols(['.'], 'test_handcalc.clsp')
test_handcalc_program = Program.from_bytes(bytes.fromhex(open('test_handcalc.clvm.hex').read()))

def test_smoke_compare():
    compare_program.run(Program.to([]))

def test_handcalc():
    diag_run_clvm(test_handcalc_program, Program.to([]), 'test_handcalc.sym')

def proper_list_inner(result,cl):
    if hasattr(cl, 'pair') and cl.pair is not None:
        result.append(cl.pair[0])
        return proper_list_inner(result,cl.pair[1])
    else:
        return result

def proper_list(cl):
    result = []
    return proper_list_inner(result,cl)

def int_list(cl):
    return [int(x) for x in Program.to(cl).as_iter()]

def de_none_list(l):
    return [x if x is not None else [] for x in l]

def with_random_lists(n,f):
    for length in range(n): # 0-10 length
        for i in range(1 + (3 * length)): # A few orders each
            orig_list = [random.randint(0,100) for x in range(length)]
            f(orig_list)

def test_prepend():
    for length1 in range(5):
        list_1 = list(range(length1))
        for length2 in range(length1):
            prepend_result = test_prepend_program.run([Program.to(list_1[:length2]),Program.to(list_1[length2:])])
            assert list_1 == int_list(prepend_result)

def test_reverse():
    def test_reverse_list(l):
        rev_args = Program.to([l])
        reversed_result = Program.to(list(reversed(l)))
        reversed_by_prog = test_reverse_program.run(rev_args)
        assert reversed_result == reversed_by_prog

    with_random_lists(10,test_reverse_list)

def test_range():
    for length in range(10):
        want_list = list(range(length))
        result = test_range_program.run(Program.to([length]))
        assert want_list == result

def do_test_permutations_of_size_n(n):
    try_list = [random.randint(0,100) for x in range(n)]
    want_set = list([list(v) for v in sorted(permutations(try_list))])
    listed_result = smoke_test_permutations_program.run(Program.to([try_list]))
    pl = proper_list(listed_result)
    perms_result = sorted([int_list(x) for x in de_none_list(pl)])
    assert want_set == perms_result

def test_permutations_0():
    do_test_permutations_of_size_n(0)

def test_permutations_1():
    do_test_permutations_of_size_n(1)

def test_permutations_2():
    n = 2
    all_a_string = 0x616161616161
    all_b_string = 0x626262626262
    for try_list in [[all_a_string,all_b_string], [all_b_string,all_a_string]]:
        want_set = list([list(v) for v in sorted(permutations(try_list))])
        listed_result = diag_run_clvm(smoke_test_permutations_program, Program.to([try_list]), 'smoke_test_permutations.sym')
        pl = proper_list(listed_result)
        perms_result = sorted([int_list(x) for x in de_none_list(pl)])
        assert want_set == perms_result

def test_chialisp_sort_program():
    diag_run_clvm(test_sort_program, Program.to([]), 'test_sort.sym')

def test_permutations_n():
    for i in range(3,6):
        do_test_permutations_of_size_n(i)

def test_chialisp_permutations_program():
    diag_run_clvm(test_permutations_program, Program.to([3, 5]), 'test_permutations.sym')

def test_smoke_sort():
    for length in range(7): # 0-7 length
        for i in range(1 + (3 * length)): # A few orders each
            orig_list = [random.randint(0,100) for x in range(length)]
            sort_args = Program.to([orig_list])
            sorted_list = Program.to(sorted(orig_list))
            sort_res = sort_program.run(sort_args)
            assert sort_res == sorted_list

if __name__ == '__main__':
    test_smoke_sort()



import binascii
import clvm_tools_rs
from pathlib import Path

classic_code = """
(mod (N X)
  (defun F (N X) (if N (sha256 (F (- N 1) X)) X))
  (F N X)
  )
"""
expected_classic = "ff02ffff01ff02ff02ffff04ff02ffff04ff05ffff04ff0bff8080808080ffff04ffff01ff02ffff03ff05ffff01ff0bffff02ff02ffff04ff02ffff04ffff11ff05ffff010180ffff04ff0bff808080808080ffff010b80ff0180ff018080"

cl23_code = """
(mod (N X)
  (include *standard-cl-23*)
  (defun F (N X) (if N (sha256 (F (- N 1) X)) X))
  (F N X)
  )
"""
expected_cl23 = "ff02ffff01ff02ff02ffff04ff02ffff04ff05ffff04ff0bff8080808080ffff04ffff01ff02ffff03ff05ffff01ff0bffff02ff02ffff04ff02ffff04ffff11ff05ffff010180ffff04ff0bff808080808080ffff010b80ff0180ff018080"

expected_deinline = "ff02ffff01ff02ff02ffff04ff02ffff04ff03ffff04ffff10ff05ffff01830f424080ff8080808080ffff04ffff01ff10ff0bff0bff0bff0bff0bff0b80ff018080"

classic_error = "(mod (X) (xxx X))"
cl23_error = "(mod (X) (include *standard-cl-23*) (+ X X1))"

compiled_code = clvm_tools_rs.compile(
    classic_code,
    ["."],
    True
)
# Classic outputs symbols to the filesystem down all routes, which likely should
# be fixed.
assert compiled_code["output"] == expected_classic

compiled_code = clvm_tools_rs.compile(
    classic_code,
    ["."]
)
assert compiled_code == expected_classic

# Verify modern compilation
compiled_code = clvm_tools_rs.compile(
    cl23_code,
    ["."],
    True
)
symbols = compiled_code["symbols"]
assert symbols["__chia__main_arguments"] == "(N X)"
assert symbols["30960d7f2ddc7188a6428a11d39a13ff70d308e6cc571ffb6ed5ec8dbe4376c0_arguments"] == "(N X)"
assert symbols["30960d7f2ddc7188a6428a11d39a13ff70d308e6cc571ffb6ed5ec8dbe4376c0"] == "F"
assert compiled_code["output"] == expected_cl23

# Check compilation with a path
test_path = Path(__file__).parent

output_file = "simple_deinline_case_23.hex"
compiled_code = clvm_tools_rs.compile_clvm(
    str(test_path / "simple_deinline_case_23.clsp"),
    output_file,
    ["."],
    True
)
assert compiled_code["symbols"]["d623cecd87575189eb1518b50cecc8944a51aa6f4bb4cf6419f70e4aa34f5a20"].startswith("letbinding")
assert compiled_code["output"] == output_file
assert open(output_file).read().strip() == expected_deinline

# Check dependency output
game_referee = Path(__file__).parent / "game-referee-in-cl23"
dependencies = clvm_tools_rs.check_dependencies(
    str(game_referee / "test_reverse.clsp"),
    [str(game_referee)]
)
assert len(dependencies) == 1
assert Path(dependencies[0]) == game_referee / 'reverse.clinc'

# Better error reporting
try:
    clvm_tools_rs.compile(classic_error, [])
    assert False
except Exception as e:
    assert e.args[0] == "error can't compile (\"xxx\" 88), unknown operator compiling (\"xxx\" 88)"

try:
    clvm_tools_rs.compile(cl23_error, [])
    assert False
except Exception as e:
    assert e.args[0] == "*inline*(1):42-*inline*(1):44: Unbound use of X1 as a variable name"



from clvm_tools_rs import binutils
from clvm_tools.binutils import assemble, disassemble
from clvm_rs.program import Program
from random import randint

test_cases = [
    "(q)",
    "()",
    "(1 . 2)",
    "(() . ())",
    "(1 2 3 . ())",
    "(1 2 3)",
    "((4 5 6) (7 8 9))",
    "(i (q . 1) (q . 2) (q . 3))",
    "((0x33 . i))",
    "(g . (138935604050210 ()))",
    "0x0001",
    "0x00",
    "0x5c4859"
]

def extend_atom():
    n_choice = randint(0,14)
    if n_choice > 10:
        return ""
    return f"{chr(n_choice+97)}@"

def pad(n,v,s):
    if len(s) < n:
        return "".join(([v]*(n - len(s)))) + s
    else:
        return s

def extend_hex():
    n_choice = randint(0,400)
    if n_choice > 255:
        return ""
    return f"{pad(2,'0',hex(n_choice)[2:])}^"

def any_item():
    keys = list(filter(lambda k: k not in "&^", production_rules.keys()))
    key_choice = randint(0,len(keys) - 1)
    return keys[key_choice]

production_rules = {
    "!": lambda: "()",
    "~": lambda: "(*&",
    "&": lambda: [")", " *&", " . *)"][randint(0,2)],
    "#": lambda: str(randint(0,2**48) - (2**47)),
    "@": lambda: f"{chr(randint(0,10)+97)}{extend_atom()}",
    "$": lambda: f"0x{pad(2,'0',hex(randint(0,255))[2:])}^",
    "^": extend_hex,
    "*": any_item
}
production_keys = "".join(production_rules.keys())

def do_replacement(x):
    if x in production_rules:
        return production_rules[x]()
    else:
        return x

def generate_random_sexp(s):
    def generate_round(s):
        return "".join([do_replacement(x) for x in s])

    s = generate_round(s)
    while any([x in production_keys for x in s]):
        s = generate_round(s)

    return s

for i in range(20):
    test_cases.append(generate_random_sexp("*"))

for t in test_cases:
    print(t)
    expected = assemble(t)
    assembled = binutils.assemble_generic(lambda x,y: Program.to((x,y)), Program.to, t)
    print(expected, assembled)

    assert expected == assembled

    disassembled = binutils.disassemble_generic(bytes(assembled))
    expected_dis = disassemble(expected)
    assert expected_dis == disassembled



from __future__ import annotations

import io
from typing import Any, Callable, Dict, Set, Tuple

from chia_rs import run_chia_program, tree_hash
from clvm import SExp
from clvm.casts import int_from_bytes
from clvm.EvalError import EvalError
from clvm.serialize import sexp_from_stream, sexp_to_stream

from chia.types.blockchain_format.sized_bytes import bytes32
from chia.util.byte_types import hexstr_to_bytes
from chia.util.hash import std_hash

from .tree_hash import sha256_treehash

INFINITE_COST = 11000000000


class Program(SExp):
    """
    A thin wrapper around s-expression data intended to be invoked with "eval".
    """

    @classmethod
    def parse(cls, f) -> "Program":
        return sexp_from_stream(f, cls.to)

    def stream(self, f):
        sexp_to_stream(self, f)

    @classmethod
    def from_bytes(cls, blob: bytes) -> Program:
        # this runs the program "1", which just returns the first argument.
        # the first argument is the buffer we want to parse. This effectively
        # leverages the rust parser and LazyNode, making it a lot faster to
        # parse serialized programs into a python compatible structure
        cost, ret = run_chia_program(
            b"\x01",
            blob,
            50,
            0,
        )
        return Program.to(ret)

    @classmethod
    def fromhex(cls, hexstr: str) -> "Program":
        return cls.from_bytes(hexstr_to_bytes(hexstr))

    def __bytes__(self) -> bytes:
        f = io.BytesIO()
        self.stream(f)  # noqa
        return f.getvalue()

    def __str__(self) -> str:
        return bytes(self).hex()

    def at(self, position: str) -> "Program":
        """
        Take a string of only `f` and `r` characters and follow the corresponding path.

        Example:

        `assert Program.to(17) == Program.to([10, 20, 30, [15, 17], 40, 50]).at("rrrfrf")`

        """
        v = self
        for c in position.lower():
            if c == "f":
                v = v.first()
            elif c == "r":
                v = v.rest()
            else:
                raise ValueError(f"`at` got illegal character `{c}`. Only `f` & `r` allowed")
        return v

    def replace(self, **kwargs) -> "Program":
        """
        Create a new program replacing the given paths (using `at` syntax).
        Example:
        ```
        >>> p1 = Program.to([100, 200, 300])
        >>> print(p1.replace(f=105) == Program.to([105, 200, 300]))
        True
        >>> print(p1.replace(rrf=[301, 302]) == Program.to([100, 200, [301, 302]]))
        True
        >>> print(p1.replace(f=105, rrf=[301, 302]) == Program.to([105, 200, [301, 302]]))
        True
        ```

        This is a convenience method intended for use in the wallet or command-line hacks where
        it would be easier to morph elements of an existing clvm object tree than to rebuild
        one from scratch.

        Note that `Program` objects are immutable. This function returns a new object; the
        original is left as-is.
        """
        return _sexp_replace(self, self.to, **kwargs)

    def get_tree_hash_precalc(self, *args: bytes32) -> bytes32:
        """
        Any values in `args` that appear in the tree
        are presumed to have been hashed already.
        """
        return sha256_treehash(self, set(args))

    def get_tree_hash(self) -> bytes32:
        return bytes32(tree_hash(bytes(self)))

    def run_with_cost(self, max_cost: int, args) -> Tuple[int, "Program"]:
        prog_args = Program.to(args)
        cost, r = run_chia_program(self.as_bin(), prog_args.as_bin(), max_cost, 0)
        return cost, Program.to(r)

    def run(self, args) -> "Program":
        cost, r = self.run_with_cost(INFINITE_COST, args)
        return r

    # Replicates the curry function from clvm_tools, taking advantage of *args
    # being a list.  We iterate through args in reverse building the code to
    # create a clvm list.
    #
    # Given arguments to a function addressable by the '1' reference in clvm
    #
    # fixed_args = 1
    #
    # Each arg is prepended as fixed_args = (c (q . arg) fixed_args)
    #
    # The resulting argument list is interpreted with apply (2)
    #
    # (2 (1 . self) rest)
    #
    # Resulting in a function which places its own arguments after those
    # curried in in the form of a proper list.
    def curry(self, *args) -> "Program":
        fixed_args: Any = 1
        for arg in reversed(args):
            fixed_args = [4, (1, arg), fixed_args]
        return Program.to([2, (1, self), fixed_args])

    def uncurry(self) -> Tuple[Program, Program]:
        def match(o: SExp, expected: bytes) -> None:
            if o.atom != expected:
                raise ValueError(f"expected: {expected.hex()}")

        try:
            # (2 (1 . <mod>) <args>)
            ev, quoted_inner, args_list = self.as_iter()
            match(ev, b"\x02")
            match(quoted_inner.pair[0], b"\x01")
            mod = quoted_inner.pair[1]
            args = []
            while args_list.pair is not None:
                # (4 (1 . <arg>) <rest>)
                cons, quoted_arg, rest = args_list.as_iter()
                match(cons, b"\x04")
                match(quoted_arg.pair[0], b"\x01")
                args.append(quoted_arg.pair[1])
                args_list = rest
            match(args_list, b"\x01")
            return Program.to(mod), Program.to(args)
        except ValueError:  # too many values to unpack
            # when unpacking as_iter()
            # or when a match() fails
            return self, self.to(0)
        except TypeError:  # NoneType not subscriptable
            # when an object is not a pair or atom as expected
            return self, self.to(0)
        except EvalError:  # first of non-cons
            # when as_iter() fails
            return self, self.to(0)

    def as_int(self) -> int:
        return int_from_bytes(self.as_atom())

    def __deepcopy__(self, memo):
        return type(self).from_bytes(bytes(self))

    EvalError = EvalError


def _tree_hash(node: SExp, precalculated: Set[bytes32]) -> bytes32:
    """
    Hash values in `precalculated` are presumed to have been hashed already.
    """
    if node.listp():
        left = _tree_hash(node.first(), precalculated)
        right = _tree_hash(node.rest(), precalculated)
        s = b"\2" + left + right
    else:
        atom = node.as_atom()
        if atom in precalculated:
            return bytes32(atom)
        s = b"\1" + atom
    return bytes32(std_hash(s))


NIL = Program.from_bytes(b"\x80")


def _sexp_replace(sexp: SExp, to_sexp: Callable[[Any], SExp], **kwargs) -> SExp:
    # if `kwargs == {}` then `return sexp` unchanged
    if len(kwargs) == 0:
        return sexp

    if "" in kwargs:
        if len(kwargs) > 1:
            raise ValueError("conflicting paths")
        return kwargs[""]

    # we've confirmed that no `kwargs` is the empty string.
    # Now split `kwargs` into two groups: those
    # that start with `f` and those that start with `r`

    args_by_prefix: Dict[str, SExp] = {}
    for k, v in kwargs.items():
        c = k[0]
        if c not in "fr":
            raise ValueError("bad path containing %s: must only contain `f` and `r`")
        args_by_prefix.setdefault(c, dict())[k[1:]] = v

    pair = sexp.pair
    if pair is None:
        raise ValueError("path into atom")

    # recurse down the tree
    new_f = _sexp_replace(pair[0], to_sexp, **args_by_prefix.get("f", {}))
    new_r = _sexp_replace(pair[1], to_sexp, **args_by_prefix.get("r", {}))

    return to_sexp((new_f, new_r))



from typing import List
import binascii
import json
import os
from pathlib import Path
from clvm_tools_rs import start_clvm_program, compose_run_function, compile_clvm
from clvm_tools.binutils import assemble
from clvm_rs import Program

def compile_module_with_symbols(include_paths: List[Path], source: Path):
    path_obj = Path(source)
    file_path = path_obj.parent
    file_stem = path_obj.stem
    target_file = file_path / (file_stem + ".clvm.hex")
    sym_file = file_path / (file_stem + ".sym")
    # print(f"compile_clvm({path_obj.absolute()}, {str(target_file.absolute().as_posix())}, {include_paths}, True)")
    compile_result = compile_clvm(
        str(path_obj.resolve()), str(target_file.absolute()), [str(p) for p in include_paths], True
    )
    print(f"Writing to {target_file} {compile_result}")
    symbols = compile_result["symbols"]
    if len(symbols) != 0:
        with open(str(sym_file.absolute()), "w") as symfile:
            symfile.write(json.dumps(symbols))


def run_until_end(p):
    last = None
    location = None

    while not p.is_ended():
        step_result = p.step()
        if step_result is not None:
            last = step_result
            if 'Print' in last:
                print(f"{last['Print']}")

    return last

def diag_run_clvm(program, args, symbols, options=None):
    if options is None:
        options = {}
    hex_form_of_program = binascii.hexlify(bytes(program)).decode('utf8')
    hex_form_of_args = binascii.hexlify(bytes(args)).decode('utf8')
    symbols = json.loads(open(symbols).read())
    p = start_clvm_program(hex_form_of_program, hex_form_of_args, symbols, None, options)
    report = run_until_end(p)
    if 'Failure' in report:
        print(report['Failure'])
    else:
        return assemble(report['Final'])

if __name__ == '__main__':
    # smoke test
    import sys
    import argparse
    import traceback

    parser = argparse.ArgumentParser()
    parser.add_argument('-p', '--print', action='store_true', default=True)
    parser.add_argument('program')
    parser.add_argument('env')
    parser.add_argument('symbols')
    args = parser.parse_args()

    program = Program.fromhex(open(args.program).read())
    env = Program.fromhex(open(args.env).read())
    options = { 'print': args.print }
    diag_run_clvm(program, env, args.symbols, options)



import os
from pathlib import Path
import subprocess
import traceback

FULL_NODE='chia/full_node/puzzles'
CAT_WALLET='chia/wallet/cat_wallet/puzzles'
DID_WALLET='chia/wallet/did_wallet/puzzles'
NFT_WALLET='chia/wallet/nft_wallet/puzzles'
POOLS='chia/pools/puzzles'
CONSENSUS='chia/consensus/puzzles'
GENTEST='tests/generator/puzzles'

def full_node(x):
    return {'fname': x, 'dirname': FULL_NODE}

def cat_wallet(x):
    return {'fname': x, 'dirname': CAT_WALLET}

def did_wallet(x):
    return {'fname': x, 'dirname': DID_WALLET}

def nft_wallet(x):
    return {'fname': x, 'dirname': NFT_WALLET}

def pools(x):
    return {'fname': x, 'dirname': POOLS}

def consensus(x):
    return {'fname': x, 'dirname': CONSENSUS}

def gentest(x):
    return {'fname': x, 'dirname': GENTEST}

recompile_list = [
    full_node('block_program_zero.clsp'),
    full_node('decompress_coin_spend_entry.clsp'),
    full_node('decompress_coin_spend_entry_with_prefix.clsp'),
    full_node('decompress_puzzle.clsp'),
    cat_wallet('delegated_tail.clsp'),
    cat_wallet('everything_with_signature.clsp'),
    cat_wallet('genesis_by_coin_id.clsp'),
    cat_wallet('genesis_by_puzzle_hash.clsp'),
    did_wallet('did_innerpuz.clsp'),
    nft_wallet('nft_metadata_updater_default.clsp'),
    nft_wallet('nft_metadata_updater_updateable.clsp'),
    nft_wallet('nft_ownership_layer.clsp'),
    nft_wallet('nft_ownership_transfer_program_one_way_claim_with_royalties.clsp'),
    nft_wallet('nft_state_layer.clsp'),
    pools('pool_member_innerpuz.clsp'),
    pools('pool_waitingroom_innerpuz.clsp'),
    consensus('rom_bootstrap_generator.clsp'),
    'p2_conditions.clsp',
    'p2_delegated_conditions.clsp',
    'p2_delegated_puzzle.clsp',
    'p2_delegated_puzzle_or_hidden_puzzle.clsp',
    'p2_m_of_n_delegate_direct.clsp',
    'p2_puzzle_hash.clsp',
    'p2_singleton.clsp',
    'p2_singleton_or_delayed_puzhash.clsp',
    'settlement_payments.clsp',
    'singleton_launcher.clsp',
    'singleton_top_layer.clsp',
    'singleton_top_layer_v1_1.clsp',
    gentest('test_generator_deserialize.clsp'),
    gentest('test_multiple_generator_input_arguments.clsp')
]

for recompile_entry in recompile_list:
    if 'dirname' in recompile_entry and 'fname' in recompile_entry:
        dirname = recompile_entry['dirname']
        filename = recompile_entry['fname']
    else:
        filename = recompile_entry
        dirname = 'chia/wallet/puzzles'

    srcfile = str(Path(dirname) / Path(filename))
    hexfile = f'{str(srcfile)}.hex'
    hexdata = open(hexfile).read().strip()
    os.unlink(hexfile)
    try:
        compiled = subprocess.check_output(['../target/release/run', '-i', dirname, '-i', 'chia/wallet/puzzles', srcfile]).strip()
        recompile = subprocess.check_output(['../target/release/opc', compiled]).decode('utf8').strip()
    except:
        print(f'compiling {fname}')
        traceback.print_exc()

    if hexdata != recompile:
        print(f'*** COMPILE RESULTED IN DIFFERENT OUTPUT FOR FILE {fname}')
        assert hexdata == recompile
